from pathlib import PurePath
import subprocess
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Literal,
    Protocol,
    Type,
    TypeGuard,
    Union,
    overload,
)


if TYPE_CHECKING:
    from anyio.abc import AnyByteReceiveStream, AnyByteSendStream


StreamName = Literal["stdin", "stdout", "stderr"]
SyncIO = tuple[IO[bytes] | None, StreamName]
AsyncIO = (
    tuple[Union[None, "AnyByteReceiveStream"], Literal["stdout", "stderr"]]
    | tuple[Union[None, "AnyByteSendStream"], Literal["stdin"]]
)


class StreamError(RuntimeError):
    pass


class Stream:
    def setup(self, stream: StreamName) -> int | IO | None:
        """Will be called as Popen arg: (Popen(..., stdin=Stream.setup("stdin")))"""

    async def setup_async(self, stream: StreamName) -> int | IO | None:
        return self.setup(stream)

    def init(self, io: SyncIO):
        """Will be after process created: Stream.init(process.stdin, "stdin")"""

    async def init_async(self, io: AsyncIO):
        """Will be after process created: await Stream.init_async(process.stdin, "stdin")"""
        self.init((None, io[1]))

    def close(self):
        """Will be called after process end"""

    async def close_async(self):
        """Will be called after process end"""

    async def process_async(self):
        """Process something while process is running"""


class DescriptorReference(Stream):
    """Container for descriptor"""

    def __init__(self, descriptor: int | None = None):
        self.descriptor = descriptor

    def setup(self, stream: StreamName) -> int | IO | None:  # NOSONAR
        """Will be called as Popen arg: (Popen(..., stdin=Stream.setup("stdin")))"""
        return self.descriptor

    def init(self, io: SyncIO):
        _io, stream = io
        if has_fileno_guard(_io):
            if (
                self.descriptor
                and _io.fileno() != self.descriptor
                and self.descriptor > 0
            ):
                raise StreamError(f"Descriptor should not be overridden, {stream}")

            self.descriptor = _io.fileno()
        else:
            raise StreamError(f"Unable to get descriptor from {_io}, {stream}")

    async def init_async(self, io: AsyncIO):
        raise NotImplementedError()


class Pipe(DescriptorReference):
    """Ensures that producer process starts before consumer"""

    def setup(self, stream: StreamName) -> int:
        if stream == "stdin":
            if self.descriptor is None:
                raise StreamError(f"No descriptor to pass into {stream}")
            return self.descriptor
        else:
            if self.descriptor is not None:
                raise StreamError(
                    f"Descriptor should be empty before passing into {stream}"
                )
            return subprocess.PIPE

    async def setup_async(self, stream: StreamName):
        return subprocess.PIPE

    async def init_async(self, io: AsyncIO):
        if io[1] == "stdin":
            assert io[0] is not None
            self.async_write = io[0]
            return
        assert io[0] is not None
        self.async_read = io[0]

    async def process_async(self):
        if (
            hasattr(self, "_running")
            or not hasattr(self, "async_write")
            or not hasattr(self, "async_read")
        ):
            return
        self._running = True
        async with self.async_read, self.async_write:
            while True:
                import anyio

                try:
                    value = await self.async_read.receive()
                except anyio.EndOfStream:
                    break
                await self.async_write.send(value)


class DevNull(DescriptorReference):
    def __init__(self):
        super().__init__(subprocess.DEVNULL)

    def init(self, io: SyncIO):
        return

    async def init_async(self, io: AsyncIO):
        return


class HasFileno(Protocol):
    def fileno(self) -> int: ...


def has_fileno_guard(val: Any) -> TypeGuard[HasFileno]:
    return hasattr(val, "fileno")


class IOStream(DescriptorReference):
    sync_io: IO[bytes]

    def __init__(self):
        super().__init__(subprocess.PIPE)

    def get_io(self) -> IO[bytes]:
        return self.sync_io

    def init(self, io: SyncIO):
        assert io[0] is not None, "No stream passed"
        self.sync_io = io[0]

    async def init_async(self, io: AsyncIO):
        if io[1] == "stdin":
            assert io[0] is not None, "No async stream passed"
            self.async_write = io[0]
        else:
            assert io[0] is not None, "No async stream passed"
            self.async_read = io[0]


class Capture[T: str | bytes](IOStream):
    _output: Type[T] = bytes  # type: ignore
    data: bytes

    @overload
    def get(self: "Capture[str]") -> str: ...
    @overload
    def get(self) -> bytes: ...
    def get(self):
        if self._output is str:
            return self.data.decode()
        return self.data

    def __class_getitem__(cls, item: type[str]):
        return type(f"{cls.__name__}[{item.__name__}]", (cls,), {"_output": item})

    def close(self):
        if hasattr(self, "sync_io"):
            self.data = self.sync_io.read()

    async def close_async(self):
        if not hasattr(self, "async_read"):
            return
        self.data = await self.async_read.receive()


class Send(IOStream):
    """Write data and close"""

    def __init__(self, data: bytes | str) -> None:
        super().__init__()
        if isinstance(data, str):
            data = data.encode()
        self.data = data

    def init(self, io: SyncIO):
        super().init(io)
        self.sync_io.write(self.data)
        self.sync_io.close()

    async def init_async(self, io: AsyncIO):
        await super().init_async(io)
        await self.async_write.send(self.data)
        await self.async_write.aclose()


class FileStream(Stream):
    def __init__(self, path: PurePath | str, append: bool = True) -> None:
        self.path = path
        self.append = append
        super().__init__()

    def setup(self, stream: StreamName) -> int | IO | None:
        if stream == "stdin":
            self.file = open(self.path, "rb")
            return self.file
        self.file = open(self.path, "ab" if self.append else "wb")
        return self.file

    def close(self):
        if not self.file.closed:
            self.file.close()



from typing import Callable, overload
from .command import Command
from .patcher import patch_function
from .shell_patch import patch_shell_arguments


@overload
def sh[C: Callable](cmd: C) -> C: ...
@overload
def sh(cmd: str | list[str]) -> Command[None, None, None]: ...
def sh(cmd: str | list[str] | Callable):
    if isinstance(cmd, str | list):
        return Command(cmd)
    patch_function(cmd, patcher=patch_shell_arguments)
    return cmd


def shell(cmd: str | list[str]):
    assert isinstance(cmd, list), "apply @sh to parent function"
    return cmd



from pathlib import Path
import sys
from tempfile import TemporaryDirectory

from recmd.executor.subprocess import SubprocessExecutor
from recmd.shell import sh
from recmd.stream import Capture, IOStream, Send


@sh
def python(code: str):
    return sh(f"{sys.executable} -c {code}")


@sh
def test_process_start():
    with SubprocessExecutor().use():
        assert ~python("print()")


@sh
def test_process_stdout():
    with SubprocessExecutor().use():
        assert (~python("print(123)").output()).strip() == "123"


@sh
def test_process_stdout_file():
    with SubprocessExecutor(True).use(), TemporaryDirectory() as dir:
        assert python("print(123)") >> (Path(dir) / "test")
        assert (Path(dir) / "test").read_text() == "123\n"


@sh
def test_process_stdin():
    with SubprocessExecutor().use():
        out = ~python("print(input(), end='')").send("123").output()
        assert out == "123"


@sh
def test_process_stdin_file():
    with SubprocessExecutor().use(), TemporaryDirectory() as dir:
        dir = Path(dir)
        file = dir / "test"
        file.write_text("value")
        with (
            file >> python("import sys;print(sys.stdin.read(),end='')") >> IOStream()
        ) as process:
            result = process.stdout.sync_io.read()
            assert result == b"value"
        assert process


@sh
def test_process_stdin_bytes():
    with SubprocessExecutor().use():
        process = (
            Send("value")
            >> python("import sys;print(sys.stdin.read(),end='')")
            >> Capture[bytes]()
        )
        assert ~process
        result = process.stdout.get()
        assert result == b"value"


@sh
def test_process_pipe():
    with SubprocessExecutor().use():
        with python("print(123)") | python("print(input())") >> IOStream() as group:
            result = group.commands[-1].stdout.sync_io.read().replace(b"\r", b"")
            assert b"123\n" == result



from pathlib import Path
import sys
from tempfile import TemporaryDirectory

import anyio
import pytest

from recmd.executor.anyio import AnyioExecutor
from recmd.shell import sh
from recmd.stream import Capture, IOStream, Send


@sh
def python(code: str):
    return sh(f"{sys.executable} -c {code}")


async def receive_all(stream) -> bytes:
    result = b""
    while True:
        try:
            value = await stream.receive()
            if isinstance(value, bytearray):
                value = bytes(value)
            result += value
        except anyio.EndOfStream:
            break
    return result


@pytest.mark.anyio
@sh
async def test_process_start():
    with AnyioExecutor().use():
        assert await python("print()")


@pytest.mark.anyio
@sh
async def test_process_stdout():
    with AnyioExecutor().use():
        assert (await python("print(123)").output()).strip() == "123"


@pytest.mark.anyio
@sh
async def test_process_stdout_file():
    with AnyioExecutor().use(), TemporaryDirectory() as dir:
        assert await (python("print(123)") >> (Path(dir) / "test"))
        assert (Path(dir) / "test").read_text() == "123\n"


@pytest.mark.anyio
@sh
async def test_process_stdin():
    with AnyioExecutor().use():
        out = await python("print(input(), end='')").send("123").output()
        assert out == "123"


@pytest.mark.anyio
@sh
async def test_process_stdin_file():
    with AnyioExecutor().use(), TemporaryDirectory() as dir:
        dir = Path(dir)
        file = dir / "test"
        file.write_text("value")
        async with (
            file >> python("import sys;print(sys.stdin.read(),end='')") >> IOStream()
        ) as process:
            result = await process.stdout.async_read.receive()
            assert result == b"value"
        assert process


@pytest.mark.anyio
@sh
async def test_process_stdin_bytes():
    with AnyioExecutor().use():
        process = (
            Send("value")
            >> python("import sys;print(sys.stdin.read(),end='')")
            >> Capture[bytes]()
        )
        assert await process
        result = process.stdout.get()
        assert result == b"value"


@pytest.mark.anyio
@sh
async def test_process_pipe():
    with AnyioExecutor().use():
        async with python("print(123)") | python(
            "print(input())"
        ) >> IOStream() as group:
            result = (await receive_all(group.commands[-1].stdout.async_read)).replace(
                b"\r", b""
            )
            assert b"123\n" == result



from .exceptions import TransformError
from .patcher import (
    patch_function,
    apply_patch,
    AnyFunctionDef,
    line_attributes,
    get_ast,
)
from .shell_patch import patch_shell_arguments
from .executor.abc import SyncExecutor, AsyncExecutor
from .shell import sh, shell
from .stream import Capture, DevNull, FileStream, IOStream, Send, Stream, Pipe


__all__ = [
    "TransformError",
    "patch_function",
    "apply_patch",
    "AnyFunctionDef",
    "line_attributes",
    "get_ast",
    "patch_shell_arguments",
    "SyncExecutor",
    "AsyncExecutor",
    "sh",
    "shell",
    "Capture",
    "DevNull",
    "FileStream",
    "IOStream",
    "Send",
    "Stream",
    "Pipe",
]



import pytest
from recmd import TransformError, apply_patch, patch_shell_arguments


def shell(data: str | list[str]) -> list:
    assert isinstance(data, list)
    return data


def test_basic_split():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell("a b c") == ["a", "b", "c"]

    test()


def test_include():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(f"a b {1}") == ["a", "b", "1"]
        assert shell(f"a b {'1 2'}") == ["a", "b", "1 2"]

    test()


def test_prefix_and_postfix():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(f"a b --arg={1}postfix f") == ["a", "b", "--arg=1postfix", "f"]

    test()


def test_double_space():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(f"a b {1}  f") == ["a", "b", "1", "", "f"]

    test()


def test_quotation():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(r"a\ b c") == ["a b", "c"]
        assert shell(rf"'a b\\ {1}'") == ["a b\\ 1"]
        assert shell(rf"'a b\'{1}'") == ["a b'1"]
        assert shell(rf'"a b\"{1}"') == ['a b"1']
        assert shell(rf'"a b\"{1}" " " \ \ ') == ['a b"1', " ", "  "]

    test()


def test_unwrap_conversion():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(f"a b {[1, 2, 3]:*!s}") == ["a", "b", "1", "2", "3"]

    test()


def test_unwrap_format():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(f"a b {[1, 2, 3]:*:.1f}") == ["a", "b", "1.0", "2.0", "3.0"]

    test()


def test_unwrap_format_fallback():
    @apply_patch(patcher=patch_shell_arguments)
    def test():
        assert shell(f"a b {1:*>2}") == ["a", "b", "*1"]

    test()


def test_invalid_escape():
    with pytest.raises(TransformError):

        @apply_patch(patcher=patch_shell_arguments)  # type: ignore
        def _():
            shell("\\a")


def test_non_strict_unwrap():
    @apply_patch(patcher=patch_shell_arguments)  # type: ignore
    def test():
        assert shell(f"a b {[1, 2, 3]:*}") == ["a", "b", 1, 2, 3]

    test()


def test_unwrap_conversion_error():
    with pytest.raises(TransformError):

        @apply_patch(patcher=patch_shell_arguments)
        def _():
            shell(f"a b {[]!a:*}")


def test_unwrap_prefix_err():
    with pytest.raises(TransformError):

        @apply_patch(patcher=patch_shell_arguments)
        def _():
            shell(f"a b test{[]:*}")


def test_unwrap_suffix_err():
    with pytest.raises(TransformError):

        @apply_patch(patcher=patch_shell_arguments)
        def _():
            shell(f"a b {[]:*}test")



import ast
from typing import Callable, Concatenate
from .exceptions import TransformError
from .patcher import AnyFunctionDef, line_attributes

try:
    from loguru import logger  # type: ignore

    trace = logger.trace
except ImportError:
    import logging

    trace = logging.getLogger(__name__).debug


__all__ = ["patch_shell_arguments"]


NAMES = ["sh", "shell"]
STRING_START = ["'", '"']
ESCAPE = "\\"
Quotation = str | None


def iterate_arguments(data: str, quotation: Quotation):  # NOSONAR
    group = ""

    is_escape = False
    for char in data:
        if (
            char == " " and not is_escape and quotation is None
        ):  # space outside quotation
            yield group, quotation
            group = ""
            continue
        elif char == quotation and not is_escape:  # end of quoted string
            quotation = None
            continue
        elif char in STRING_START and not is_escape:  # start of quoted string
            quotation = char
            continue

        if is_escape:
            is_escape = False
            if (quotation and char not in [ESCAPE, quotation]) or (
                not quotation and char not in [ESCAPE, " "]
            ):
                raise TransformError(f"Invalid escape: '\\{char}' in {data!r}")
        elif char == ESCAPE:
            is_escape = True
        if not is_escape:
            group += char

    if is_escape is True:
        raise TransformError(f"Escape is not used in {data!r}")
    yield group, quotation


def split_joined_string(node: ast.JoinedStr):
    group: list[ast.expr] = []
    quotation = None
    for value in node.values:
        if not isinstance(value, ast.Constant):
            group.append(value)
            continue
        assert isinstance(value.value, str)

        for i, (chunk, quotation) in enumerate(
            iterate_arguments(value.value, quotation)
        ):
            if i > 0:
                yield group.copy()
                group.clear()
            group.append(
                ast.Constant(value=chunk, kind=value.kind, **line_attributes(value))
            )
    if quotation is not None:
        raise TransformError(f"{quotation} is not closed")
    if group:
        yield group


def merge_group(group: list[ast.expr]):
    if len(group) == 1 and isinstance(group[0], ast.Constant):
        return group[0]

    group = [x for x in group if not (isinstance(x, ast.Constant) and x.value == "")]
    if (
        len(group) == 1
        and isinstance(group[0], ast.FormattedValue)
        and isinstance(group[0].format_spec, ast.JoinedStr)
        and group[0].format_spec.values
        and isinstance(group[0].format_spec.values[0], ast.Constant)
        and isinstance(group[0].format_spec.values[0].value, str)
        and (
            group[0].format_spec.values[0].value == "*"
            or group[0].format_spec.values[0].value.startswith("*!")
            or group[0].format_spec.values[0].value.startswith("*:")
        )
    ):
        value = group[0].value
        if group[0].conversion != -1:
            raise TransformError(
                f":* arguments should have conversion after :* (:*!{chr(group[0].conversion)})"
            )
        fmt = group[0].format_spec.values[0].value.removeprefix("*")
        conversion = -1
        if fmt.startswith("!"):
            if fmt[1] not in ["a", "s", "r"]:
                raise TransformError(f"Invalid conversion: !{fmt[1]}")
            conversion = ord(fmt[1])
            fmt = fmt[2:]

        should_format = fmt.startswith(":")
        if conversion > 0 or should_format:
            fmt = fmt.removeprefix(":")
            # (f"{x!{conversion}:{fmt}}" for x in {value})
            value = ast.GeneratorExp(
                elt=ast.JoinedStr(
                    values=[
                        ast.FormattedValue(
                            conversion=conversion,
                            value=ast.Name(
                                id="x",
                                ctx=ast.Load(),
                                **line_attributes(group[0].format_spec),
                            ),
                            format_spec=ast.JoinedStr(
                                values=[
                                    ast.Constant(
                                        value=fmt,
                                        **line_attributes(group[0].format_spec),
                                    ),
                                    *group[0].format_spec.values[1:],
                                ],
                                **line_attributes(group[0].format_spec),
                            )
                            if should_format
                            else None,
                            **line_attributes(value),
                        )
                    ],
                    **line_attributes(value),
                ),
                generators=[
                    ast.comprehension(
                        target=ast.Name(
                            id="x",
                            ctx=ast.Store(),
                            **line_attributes(group[0].format_spec),
                        ),
                        iter=value,
                        ifs=[],
                        is_async=0,
                    )
                ],
                **line_attributes(group[0]),
            )

        return ast.Starred(value, ast.Load(), **line_attributes(group[0].value))
    max_end_line = max(node.end_lineno for node in group if node.end_lineno)
    for node in group:
        if (  # TODO: refactor?
            isinstance(node, ast.FormattedValue)
            and isinstance(node.format_spec, ast.JoinedStr)
            and node.format_spec.values
            and isinstance(node.format_spec.values[0], ast.Constant)
            and isinstance(node.format_spec.values[0].value, str)
            and (
                node.format_spec.values[0].value == "*"
                or node.format_spec.values[0].value.startswith("*!")
                or node.format_spec.values[0].value.startswith("*:")
            )
        ):
            raise TransformError(":* arguments should not have prefixes/postfixes")

    return ast.JoinedStr(
        list(group),
        lineno=min(node.lineno for node in group),
        col_offset=min(node.col_offset for node in group),
        end_col_offset=max(
            node.end_col_offset
            for node in group
            if node.end_lineno == max_end_line and node.end_col_offset
        ),
        end_lineno=max_end_line,
    )


def string_to_list(node: ast.JoinedStr | ast.Constant):
    if isinstance(node, ast.Constant):
        node = ast.JoinedStr(values=[node], **line_attributes(node))
    arguments = []
    for group in split_joined_string(node):
        arguments.append(merge_group(group))

    return ast.List(
        elts=arguments,
        ctx=ast.Load(),
        **line_attributes(node),
    )


def patch_shell_arguments[**P](
    function: AnyFunctionDef,
    names: list[str] = NAMES,
    convert: Callable[
        Concatenate[ast.JoinedStr | ast.Constant, P],
        ast.expr,
    ] = string_to_list,
    *args: P.args,
    **kwargs: P.kwargs,
):
    r"""
    Transform f-string argument to list that can be safely passed to shell

    rules:
    1.  only calls with signature `sh("")` / `sh(f"")` / `shell(f"")` / `something.sh(f"")` / `something.shell(f"")` are changed
    2.  split argument 0 by space (`"a b"` -> `["a", "b"]`)
    3.  double space are honored (`"a  b"` -> `["a", "", "b"]`)
    4.  f-strings arguments are inlined within respected arguments (f"abc{'def'}gh 123{456}" -> [f"abd{'def'}gh", f"123{456}"])
    5.  arguments with :* formatting are unwrapped ("ls -l {['dir', 'dir2']!a}" -> ["ls", "-l", "dir", "dir2"])
    6.  formatting for :* arguments can be specified after :* ("ls -l {[1]:*:.2f}" -> ["ls", "-l", "1.00"])
    7.  conversion for :* arguments also can be specified after :* but before formatting ("ls -l {[1]:*!s}" -> ["ls", "-l", "1"])
    8.  :* is passed as is if no formatting applied ("ls -l {[1]:*}" -> ["ls", "-l", 1])
    9.  arguments can be quoted to include spaces ("'a b' c" -> ["a b", "c"])
    10. spaces and quotes can be escaped using \ ("'a\' b' c d\ e \\" -> ["a' b", "c", "d e", "\"])

    Examples:
    * `shell(f'echo --arg={1}postfix')` to `shell([f'echo', f'--arg={1}postfix'])`
    * `shell(f'echo {['a', 'b']:*}')` to `shell([f'echo', *['a', 'b']])`
    """

    for node in ast.walk(function):
        if (
            isinstance(node, ast.Call)
            and (
                # __sh(...)__
                (isinstance(node.func, ast.Name) and node.func.id in names)
                # __object.sh(...)__
                or (isinstance(node.func, ast.Attribute) and node.func.attr in names)
            )
            and (
                # ...(__f""__)
                isinstance(node.args[0], ast.JoinedStr)
                # ...(__""__)
                or (
                    isinstance(node.args[0], ast.Constant)
                    and isinstance(node.args[0].value, str)
                )
            )
        ):
            prev = ast.unparse(node)
            node.args[0] = convert(node.args[0], *args, **kwargs)
            trace(f"transformed shell {prev} -> {ast.unparse(node)}")



from abc import ABC, abstractmethod
from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, AsyncContextManager, ClassVar, ContextManager


if TYPE_CHECKING:
    from recmd.command import Command


class AsyncExecutor(ABC):
    context: ClassVar = ContextVar["AsyncExecutor"]("recmd.executor.AsyncExecutor")

    @classmethod
    def get(cls) -> "AsyncExecutor":
        self = cls.context.get(None)
        if self is None:
            raise RuntimeError(
                "No async executor is set (to enable default install anyio)"
            )
        return self

    @abstractmethod
    def run(self, command: "Command") -> AsyncContextManager: ...

    @contextmanager
    def use(self):
        reset = self.context.set(self)
        try:
            yield self
        finally:
            self.context.reset(reset)


class SyncExecutor(ABC):
    context: ClassVar = ContextVar["SyncExecutor"]("recmd.executor.SyncExecutor")

    implicit_start: bool
    """Allow process to be started via assert"""

    def __init__(self, implicit_start: bool = False) -> None:
        super().__init__()

        self.implicit_start = implicit_start

    @classmethod
    def get(cls) -> "SyncExecutor":
        self = cls.context.get(None)
        if self is None:
            raise RuntimeError("No sync executor is set")
        return self

    @abstractmethod
    def run(self, command: "Command") -> ContextManager: ...

    @contextmanager
    def use(self):
        reset = self.context.set(self)
        try:
            yield self
        finally:
            self.context.reset(reset)



from contextlib import ExitStack, contextmanager
import os
from pathlib import PurePath
from subprocess import Popen
from typing import IO
from recmd.command import AnyStream, Command, CompleteCommand, RunningCommand
from recmd.executor.abc import SyncExecutor
from recmd.stream import FileStream, Stream, StreamName


class SubprocessExecutor(SyncExecutor):
    @contextmanager
    def run(self, command: Command):
        with ExitStack() as stack:
            stdin, stdin_stream = self.prepare_stream(command.stdin, "stdin")
            stack.callback(self.close_stream, stdin_stream)
            stdout, stdout_stream = self.prepare_stream(command.stdout, "stdout")
            stack.callback(self.close_stream, stdout_stream)
            stderr, stderr_stream = self.prepare_stream(command.stderr, "stderr")
            stack.callback(self.close_stream, stderr_stream)

            env = command.environment
            if command.inherit_env:
                env = os.environ | env
            process = Popen(
                command.cmd,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                env=env,
                cwd=command.cwd,
            )
            command.running = RunningCommand(process.pid, process)

            self.setup_stream(stdin_stream, process.stdin, "stdin")
            self.setup_stream(stdout_stream, process.stdout, "stdout")
            self.setup_stream(stderr_stream, process.stderr, "stderr")
            try:
                yield
            finally:
                status = process.wait()
                command.complete = CompleteCommand(status)

    def setup_stream(
        self, stream: Stream | None, io: IO[bytes] | None, name: StreamName
    ):
        if stream is None or io is None:
            return
        stream.init((io, name))

    def close_stream(self, stream: Stream | None):
        if stream is None:
            return
        stream.close()

    def prepare_stream(self, stream: AnyStream, name: StreamName):
        if isinstance(stream, str | PurePath):
            stream = FileStream(stream)
        if isinstance(stream, Stream):
            return stream.setup(name), stream
        return stream, None



from contextlib import AsyncExitStack, ExitStack
from pathlib import PurePath
from typing import IO, Any, Callable, Literal, Self, cast, overload

from .executor.abc import AsyncExecutor, SyncExecutor
from .map_result import ResultMapper
from .stream import Pipe, Capture, Send, Stream


AnyStream = str | int | PurePath | Stream | None | IO


class Command[IN: AnyStream, OUT: AnyStream, ERR: AnyStream]:
    running: "RunningCommand"
    """Should be populated by executor"""
    complete: "CompleteCommand"
    """Should be populated by executor"""
    stdin: IN
    stdout: OUT
    stderr: ERR

    def __init__(
        self,
        cmd: list[str] | str,
        stdin: IN = None,
        stdout: OUT = None,
        stderr: ERR = None,
        inherit_env: bool = True,
        env: dict[str, str] | None = None,
    ) -> None:
        assert isinstance(
            cmd, list
        ), "apply @sh decorator to function containing this statement"
        self.cmd = cmd
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.cwd = None
        self.inherit_env = inherit_env
        self.environment = env or {}

    def did_start(self):
        return hasattr(self, "running")

    def did_complete(self):
        return hasattr(self, "complete")

    def pid(self):
        assert self.did_start()
        return self.running.pid

    def status(self) -> ResultMapper[Self, Callable[[Self], int]]:
        return self.map().apply(lambda x: x.complete.status)

    @overload  # (Command[*, None, *])           -> ResultMapper[Command[*, Capture[str],   *], ..., str]
    def output[_PI: AnyStream, _PE: AnyStream](
        self: "Command[_PI, None, _PE]", to_string: Literal[True] = True
    ) -> ResultMapper[
        "Command[_PI,  Capture[str], _PE]",
        Callable[["Command[_PI, Capture[str], _PE]"], str],
    ]: ...
    @overload  # (Command[*, None, *], False)    -> ResultMapper[Command[*, Capture[bytes], *], ..., bytes]
    def output[_PI: AnyStream, _PE: AnyStream](
        self: "Command[_PI, None, _PE]", to_string: Literal[False]
    ) -> ResultMapper[
        "Command[_PI,  Capture[bytes], _PE]",
        Callable[["Command[_PI, Capture[bytes], _PE]"], bytes],
    ]: ...
    @overload  # (Command[*, Capture[str], *])   -> ResultMapper[Command[*, Capture[str],   *], ..., str]
    def output[_PI: AnyStream, _PE: AnyStream](
        self: "Command[_PI, Capture[str], _PE]",
    ) -> ResultMapper[
        "Command[_PI,  Capture, _PE]",
        Callable[["Command[_PI, Capture[str], _PE]"], str],
    ]: ...
    @overload  # (Command[*, Capture[bytes], *]) -> ResultMapper[Command[*, Capture[bytes], *], ..., bytes]
    def output[_PI: AnyStream, _PE: AnyStream](
        self: "Command[_PI, Capture[bytes], _PE]",
    ) -> ResultMapper[
        "Command[_PI,  Capture[bytes], _PE]",
        Callable[["Command[_PI, Capture[bytes], _PE]"], bytes],
    ]: ...
    @overload  # (Command[*, Capture[Any], *])   -> ResultMapper[Command[*, Capture[Any],   *], ..., bytes]
    def output[_PI: AnyStream, _PE: AnyStream](
        self: "Command[_PI, Capture, _PE]",
    ) -> ResultMapper[
        "Command[_PI,  Capture, _PE]", Callable[["Command[_PI, Capture, _PE]"], bytes]
    ]: ...

    def output[_PI: AnyStream, _PE: AnyStream](  # type: ignore
        self: "Command[_PI, Any, _PE]", to_string: bool = True
    ) -> ResultMapper[
        "Command[_PI,  Capture, _PE]", Callable[["Command[_PI, Capture, _PE]"], bytes]
    ]:
        if self.stdout is None:
            if to_string:
                self.stdout = Capture[str]()
            else:
                self.stdout = Capture[bytes]()
        else:
            assert isinstance(self.stdin, Capture)
        return self.map().apply(lambda x: x.stdout.get())  # type: ignore

    def send[_PO: AnyStream, _PE: AnyStream](
        self: "Command[None, _PO, _PE]", data: str | bytes
    ) -> "Command[Send, _PO, _PE]":
        assert self.stdin is None

        return self.with_stdin(Send(data))

    def run(self):
        if self.did_start():
            if self.did_complete():
                return self
            self.running.wait().run()
            return self
        with SyncExecutor.get().run(self):
            return self

    async def run_async(self):
        if self.did_start():
            if self.did_complete():
                return self
            await self.running.wait()
            return self
        async with AsyncExecutor.get().run(self):
            return self

    def map(self):
        return ResultMapper(self)

    def env(self, env: dict[str, str] = {}, **kwargs: str):
        self._assert_not_started()
        self.environment.update(env)
        self.environment.update(kwargs)
        return self

    def _assert_not_started(self):
        assert not self.did_start(), "Unable to change started command"

    def with_stdin[_PO: AnyStream, _PE: AnyStream, _NI: AnyStream](
        self: "Command[Any, _PO, _PE]", stdin: _NI
    ) -> "Command[_NI, _PO, _PE]":
        self._assert_not_started()
        self.stdin = stdin
        return self  # type: ignore

    def with_stdout[_PI: AnyStream, _PE: AnyStream, _NO: AnyStream](
        self: "Command[_PI, Any, _PE]", stdout: _NO
    ) -> "Command[_PI, _NO, _PE]":
        self._assert_not_started()
        self.stdout = stdout
        return self  # type: ignore

    def with_stderr[_PI: AnyStream, _PO: AnyStream, _NE: AnyStream](
        self: "Command[_PI, _PO, Any]", stderr: _NE
    ) -> "Command[_PI, _PO, _NE]":
        self._assert_not_started()
        self.stderr = stderr
        return self  # type: ignore

    def with_cwd(self, cwd: str | PurePath | None):
        self.cwd = cwd
        return self

    def __rshift__[_PI: AnyStream, _PE: AnyStream, _NO: AnyStream](
        self: "Command[_PI, None, _PE]", stdout: _NO
    ) -> "Command[_PI, _NO, _PE]":
        return self.with_stdout(stdout)

    def __rrshift__[_PO: AnyStream, _PE: AnyStream, _NI: AnyStream](
        self: "Command[None, _PO, _PE]", stdin: _NI
    ) -> "Command[_NI, _PO, _PE]":
        return self.with_stdin(stdin)  # type: ignore

    def __ge__[_PI: AnyStream, _PO: AnyStream, _NE: AnyStream](
        self: "Command[_PI, _PO, None]", stderr: _NE
    ) -> "Command[_PI, _PO, _NE]":
        return self.with_stderr(stderr)  # type: ignore

    def copy(self):
        copy = type(self)(self.cmd)
        copy.__dict__ = self.__dict__.copy()
        return copy

    def __invert__(self):
        return self.run()

    def __await__(self):
        return self.run_async().__await__()

    def __or__[_PI: AnyStream, _PE: AnyStream, _NO: AnyStream, _NE: AnyStream](
        self: "Command[_PI, None, _PE]", value: "Command[None, _NO, _NE]"
    ) -> "CommandGroup[Command[_PI, Pipe, _PE], Command[Pipe, _NO, _NE]]":
        pipe = Pipe()
        return CommandGroup(self.with_stdout(pipe), value.with_stdin(pipe))

    def __and__[C: "Command"](self, value: C) -> "CommandGroup[Self, C]":
        return CommandGroup(self, value)

    def __bool__(self):
        if not self.did_complete():
            executor = SyncExecutor.context.get(None)
            if executor is None or not executor.implicit_start:
                raise RuntimeError(
                    "Unable to determine process status because process did not start and implicit_start is disabled"
                )
            self.run()
        return self.complete.status == 0

    def __enter__(self):
        self._ctx = SyncExecutor.get().run(self)
        self._ctx.__enter__()

        return self

    def __exit__(self, *args):
        self._ctx.__exit__(*args)

    async def __aenter__(self):
        self._actx = AsyncExecutor.get().run(self)
        await self._actx.__aenter__()

        return self

    async def __aexit__(self, *args):
        await self._actx.__aexit__(*args)


class RunningCommand:
    def __init__(self, pid: int, process: Any) -> None:
        self.pid = pid
        self._process = process

    def kill(self):
        return ResultMapper(self._process, no_run=True).apply(lambda x: x.kill())

    def terminate(self):
        return ResultMapper(self._process, no_run=True).apply(lambda x: x.terminate())

    def poll(self):
        return ResultMapper(self._process, no_run=True).apply(
            lambda x: cast(int | None, x.poll())
        )

    def wait(self):
        return ResultMapper(self._process, no_run=True).apply(
            lambda x: cast(int | None, x.wait())
        )


class CompleteCommand:
    def __init__(self, status: int) -> None:
        self.status = status


class CommandGroup[*C]:
    def __init__(self, *commands: *C) -> None:
        self.commands = commands

    def __or__[*_C, _PI: AnyStream, _PE: AnyStream, _NO: AnyStream, _NE: AnyStream](
        self: "CommandGroup[*_C, Command[_PI, None, _PE]]",
        value: Command[None, _NO, _NE],
    ) -> "CommandGroup[*_C, Command[_PI, Pipe, _PE], Command[Pipe, _NO, _NE]]":
        pipe = Pipe()
        return CommandGroup(
            *self.commands[:-1],
            self.commands[-1].with_stdout(pipe),  # type: ignore
            value.with_stdin(pipe),
        )

    @overload
    def __and__[CMD: "Command"](self, value: CMD) -> "CommandGroup[*C, CMD]": ...
    @overload
    def __and__[*_C](self, value: "CommandGroup[*_C]") -> "CommandGroup[*C, *_C]": ...
    def __and__(self, value):
        if isinstance(value, CommandGroup):
            return CommandGroup(*self.commands, *value.commands)
        return CommandGroup(*self.commands, value)

    def run(self):
        with self:
            return self

    async def run_async(self):
        async with self:
            return self

    def __enter__(self):
        self._ctx = ExitStack()
        self._ctx.__enter__()
        for command in self.commands:
            self._ctx.enter_context(command)  # type: ignore

        return self

    def __exit__(self, *args):
        self._ctx.__exit__(*args)

    async def __aenter__(self):
        self._actx = AsyncExitStack()
        await self._actx.__aenter__()
        for command in self.commands:
            await self._actx.enter_async_context(command)  # type: ignore
        return self

    async def __aexit__(self, *args):
        await self._actx.__aexit__(*args)

    def __invert__(self):
        return self.run()

    def __await__(self):
        return self.run_async().__await__()



from contextlib import AsyncExitStack, asynccontextmanager
import os
from pathlib import PurePath

from anyio import create_task_group
import anyio.abc
from recmd.command import AnyStream, Command, CompleteCommand, RunningCommand
from recmd.executor.abc import AsyncExecutor
from recmd.stream import FileStream, Stream, StreamName, AsyncIO


class AnyioExecutor(AsyncExecutor):
    @asynccontextmanager
    async def run(self, command: Command):
        async with AsyncExitStack() as stack:
            stdin, stdin_stream = await self.prepare_stream(command.stdin, "stdin")
            stack.push_async_callback(self.close_stream, stdin_stream)
            stdout, stdout_stream = await self.prepare_stream(command.stdout, "stdout")
            stack.push_async_callback(self.close_stream, stdout_stream)
            stderr, stderr_stream = await self.prepare_stream(command.stderr, "stderr")
            stack.push_async_callback(self.close_stream, stderr_stream)

            env = command.environment
            if command.inherit_env:
                env = os.environ | env
            process = await anyio.open_process(
                command.cmd,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                env=env,
                cwd=command.cwd,
            )

            command.running = RunningCommand(process.pid, process)

            await self.setup_stream(stdin_stream, (process.stdin, "stdin"))
            await self.setup_stream(stdout_stream, (process.stdout, "stdout"))
            await self.setup_stream(stderr_stream, (process.stderr, "stderr"))

            try:
                async with create_task_group() as tg:
                    tg.start_soon(self.process_stream, stdin_stream)
                    tg.start_soon(self.process_stream, stdout_stream)
                    tg.start_soon(self.process_stream, stderr_stream)
                    yield
            finally:
                status = await process.wait()
                command.complete = CompleteCommand(status)

    async def setup_stream(self, stream: Stream | None, io: AsyncIO):
        if stream is None or io is None:
            return
        await stream.init_async(io)

    async def process_stream(self, stream: Stream | None):
        if stream is None:
            return
        await stream.process_async()

    async def close_stream(self, stream: Stream | None):
        if stream is None:
            return
        await stream.close_async()

    async def prepare_stream(self, stream: AnyStream, name: StreamName):
        if isinstance(stream, str | PurePath):
            stream = FileStream(stream)
        if isinstance(stream, Stream):
            return await stream.setup_async(name), stream
        return stream, None



import ast
import traceback
from recmd.patcher import AnyFunctionDef, apply_patch, line_attributes


def fix_addition(function: AnyFunctionDef):
    for node in ast.walk(function):
        if isinstance(node, ast.BinOp) and isinstance(node.op, ast.Add):
            node.left = ast.Constant(value=1, **line_attributes(node.left))
            node.right = ast.Constant(value=1, **line_attributes(node.right))


def fix_assert_false(function: AnyFunctionDef):
    for node in ast.walk(function):
        if (
            isinstance(node, ast.Assert)
            and isinstance(node.test, ast.Constant)
            and node.test.value is False
        ):
            node.test.value = True


def test_function_patch():
    @apply_patch(patcher=fix_addition)
    def test_function_patch():
        assert 1 + 2 == 2

    test_function_patch()


def test_multiple_patch():
    @apply_patch(fix_assert_false)
    @apply_patch(fix_addition)
    def test_multiple_patch():
        assert 1 + 2 == 2
        assert False

    test_multiple_patch()


def test_lineno():
    @apply_patch(fix_assert_false)
    @apply_patch(fix_addition)
    def test_lineno():
        assert 1 + 2 == 2
        try:
            assert 2 == 1
            raise RuntimeError("Should raise error")
        except AssertionError as e:
            frame = traceback.extract_tb(e.__traceback__)[0]
            assert frame.line == "assert 2 == 1"

        assert False

    test_lineno()



import ast
from contextlib import suppress
import inspect
import sys
import types
from typing import TYPE_CHECKING, Annotated, Any, Callable, Concatenate

try:
    from typing_extensions import Doc  # type: ignore
except ImportError:
    if not TYPE_CHECKING:

        def _doc(x):
            return x

        Doc = _doc

__all__ = [
    "patch_function",
    "apply_patch",
    "AnyFunctionDef",
    "line_attributes",
    "get_ast",
]

AnyFunctionDef = ast.FunctionDef | ast.AsyncFunctionDef
type Patcher[**P] = Annotated[
    Callable[Concatenate[AnyFunctionDef, P], Any],
    Doc("Function that accepts ast.FunctionDef and modifies it inplace"),
]


def getsource(obj):
    """inspect.getsource with PEP 302 (New Import Hooks) support"""
    if (
        hasattr(obj, "__globals__")
        and hasattr(obj, "__module__")
        and (loader := obj.__globals__.get("__loader__")) is not None
        and hasattr(loader, "get_source")
    ):
        with suppress(Exception):
            return loader.get_source(obj.__module__)
    return inspect.getsource(sys.modules[obj.__module__])


def get_ast(obj: Callable):
    """Get ast.Module containing with only function from arguments in body, will add __recmd_ast__ attribute to allow multiple patches"""
    if hasattr(obj, "__recmd_ast__"):
        assert (
            isinstance(obj.__recmd_ast__, ast.Module)
            and len(obj.__recmd_ast__.body) == 1
            and isinstance(obj.__recmd_ast__.body[0], ast.FunctionDef)
        ), "Invalid value of __recmd_ast__"
        return obj.__recmd_ast__
    tree = ast.parse(getsource(obj))
    body = None
    for item in ast.walk(tree):
        if (
            isinstance(item, AnyFunctionDef)
            and item.name == obj.__name__
            and item.end_lineno
            and (
                (
                    item.decorator_list
                    and item.decorator_list[0].lineno == obj.__code__.co_firstlineno
                )
                or item.lineno == obj.__code__.co_firstlineno
            )
        ):
            assert (
                body is None
            ), f"Unable to resolve double definition of {obj.__name__} in {inspect.getfile(obj)}"
            body = item
    assert (
        body is not None
    ), f"Unable to resolve definition of {obj.__name__} in {inspect.getfile(obj)}"
    tree.body = [body]
    obj.__recmd_ast__ = tree
    return obj.__recmd_ast__


def apply_ast(function: Callable, module: ast.Module):
    """Update __code__ of function inplace"""
    consts = compile(module, inspect.getfile(function), "exec").co_consts
    code = None
    # fix for inline generics (def name[T])
    for const in consts:
        if isinstance(const, types.CodeType) and const.co_name.startswith("<"):
            consts = consts + const.co_consts

    for const in consts:
        if (
            isinstance(const, types.CodeType)
            and const.co_name == function.__name__
            and const.co_firstlineno == function.__code__.co_firstlineno
        ):
            assert code is None, "Unable to find right code"
            code = const

    assert code is not None, "Unable to find code"
    function.__code__ = code


def patch_function(
    function: Annotated[
        Callable,
        Doc("Ast of this function will be passed to patcher and then updated in place"),
    ],
    patcher: Patcher,
):
    """Update __code__ of function inplace. Multiple patches can be applied using this function"""

    module = get_ast(function)
    assert isinstance(module.body[0], AnyFunctionDef)
    patcher(module.body[0])
    apply_ast(function, module)


def apply_patch(patcher: Patcher):
    """Update __code__ of function inplace."""

    def wrapper[F: Callable](function: F) -> F:
        patch_function(function, patcher)
        return function

    return wrapper


def line_attributes(value: ast.AST):
    """Helper function for ast manipulation"""

    return dict(
        lineno=value.lineno,
        col_offset=value.col_offset,
        end_lineno=value.end_lineno,
        end_col_offset=value.end_col_offset,
    )



from inspect import iscoroutine
from typing import (
    Any,
    Callable,
    Coroutine,
    Generator,
    Never,
    Protocol,
    Self,
    overload,
)


class Runnable(Protocol):
    def run(self) -> Self: ...

    async def run_async(self) -> Self: ...


class ResultMapper[I: Runnable, *O]:
    input: I

    def __init__(self, input: I, *steps: *O, no_run: bool = False) -> None:
        self.input = input
        self._steps = steps
        self.no_run = no_run

    @overload
    def apply[_I: Runnable, NT](
        self: "ResultMapper[_I]",
        cb: Callable[[_I], NT],
    ) -> "ResultMapper[_I, Callable[[_I], NT]]": ...
    @overload
    def apply[_I: Runnable, *_S, TI, T, NT](  # NOSONAR
        self: "ResultMapper[_I, *_S, Callable[[TI], Coroutine[Any,Any,T]]]",
        cb: Callable[[T], Coroutine[Any, Any, NT] | NT],
    ) -> "ResultMapper[_I, *_S, Callable[[TI], Coroutine[Any,Any,T]], Callable[[T], Coroutine[Any,Any,NT]]]": ...
    @overload
    def apply[_I: Runnable, *_S, TI, T, NT](  # NOSONAR
        self: "ResultMapper[_I, *_S, Callable[[TI], T]]", cb: Callable[[T], NT]
    ) -> "ResultMapper[_I, *_S, Callable[[TI], T], Callable[[T], NT]]": ...

    def apply(self, cb):  # type: ignore
        return ResultMapper(self.input, *self._steps, cb, no_run=self.no_run)

    @overload
    def run[_I: Runnable, *_S](
        self: "ResultMapper[_I, *_S, Callable[[Any], Coroutine], Any]",
    ) -> Never: ...  # warn user that that user should use run_async
    @overload
    def run[_I: Runnable, *_S, T](
        self: "ResultMapper[_I, *_S, Callable[[Any], T]]",
    ) -> T: ...
    def run[_I: Runnable, *_S, T](
        self: "ResultMapper[_I, *_S, Callable[[Any], T]]",
    ) -> T:
        if self.no_run:
            value = self.input
        else:
            value = self.input.run()
        for step in self._steps:
            value = step(value)  # type: ignore

        return value  # type: ignore

    async def run_async[_I: Runnable, *_S, T](
        self: "ResultMapper[_I, *_S, Callable[[Any], Coroutine[Any,Any,T] | T]]",
    ) -> T:
        if self.no_run:
            value = self.input
        else:
            value = await self.input.run_async()

        for step in self._steps:
            value = step(value)  # type: ignore
            if iscoroutine(value):
                value = await value

        return value  # type: ignore

    def __invert__[_I: Runnable, *_S, T](
        self: "ResultMapper[_I, *_S, Callable[[Any], T]]",
    ) -> T:
        return self.run()

    def __await__[_I: Runnable, *_S, T](
        self: "ResultMapper[_I, *_S, Callable[[Any], Coroutine[Any,Any,T] | T]]",
    ) -> Generator[Any, Any, T]:
        return self.run_async().__await__()
